# Simple parser of RSS 2.0 and Atom 1.0 feeds.
#
# Example:
#
# set root [::feed::init dom_doc "my.xml"]
#
# puts "feed type: [::feed::get_type $root]"
#
# if {[catch {feed::parse $root fdct} cause]} {
# 	puts "$::errorInfo"; exit 1
# }
# ::feed::dump $fdct
#
# $dom_doc delete; # free tDOM memory

package require myutils
package require tdom

namespace eval ::feed::u {
    # iana_charsets -> tcl_charset_names
    array set enc {
		us-ascii ascii
		utf-8 utf-8
		utf-16 unicode
		iso-8859-1 iso8859-1
		iso-8859-2 iso8859-2
		iso-8859-3 iso8859-3
		iso-8859-4 iso8859-4
		iso-8859-5 iso8859-5
		iso-8859-6 iso8859-6
		iso-8859-7 iso8859-7
		iso-8859-8 iso8859-8
		iso-8859-9 iso8859-9
		iso-8859-10 iso8859-10
		iso-8859-13 iso8859-13
		iso-8859-14 iso8859-14
		iso-8859-15 iso8859-15
		iso-8859-16 iso8859-16
		iso-2022-kr iso2022-kr
		euc-kr euc-kr
		iso-2022-jp iso2022-jp
		koi8-r koi8-r
		koi8-u koi8-u
		shift_jis shiftjis
		euc-jp euc-jp
		gb2312 gb2312
		big5 big5
		cp866 cp866
		cp1250 cp1250
		cp1253 cp1253
		cp1254 cp1254
		cp1255 cp1255
		cp1256 cp1256
		cp1257 cp1257

		windows-1251 cp1251
		cp1251 cp1251

		windows-1252 cp1252
		cp1252 cp1252

		iso_8859-1:1987 iso8859-1
		iso-ir-100 iso8859-1
		iso_8859-1 iso8859-1
		latin1 iso8859-1
		l1 iso8859-1
		ibm819 iso8859-1
		cp819 iso8859-1
		csisolatin1 iso8859-1

		iso_8859-2:1987 iso8859-2
		iso-ir-101 iso8859-2
		iso_8859-2 iso8859-2
		iso-8859-2 iso8859-2
		latin2 iso8859-2
		l2 iso8859-2
		csisolatin2 iso8859-2

		iso_8859-5:1988 iso8859-5
		iso-ir-144 iso8859-5
		iso_8859-5 iso8859-5
		iso-8859-5 iso8859-5
		cyrillic iso8859-5
		csisolatincyrillic iso8859-5

		ms_kanji shiftjis
		csshiftjis shiftjis

		csiso2022kr iso2022-kr

		ibm866 cp866
		csibm866 cp866
    }
}

proc ::feed::u::iana2tcl {iana} {
	variable enc

	set iana [string tolower $iana]
	if {[info exists enc($iana)]} { return $enc($iana) }
	return ""
}

# filename -- XML file
#
# Return: an encoding name or "utf-8" if encoding wasn't found.
#
# Side effects:
# * i/o errors
# * stderr warning if encoding wasn't found
proc ::feed::u::getEncoding { filename } {
	set enc "utf-8"

	set fd [open $filename]
	fconfigure $fd -encoding binary -translation binary
	set bytes [read $fd 1024]
	close $fd

	if {[regexp {<\?xml[^>]+encoding=["']([^\"']*)["']} $bytes match e] } {
		if {[set e [iana2tcl $e]] != ""} {
			set enc $e
		} else {
			myutils::warnx "[file tail $filename]: unknown encoding"
		}
	} else {
		myutils::warnx "[file tail $filename]: no encoding specified"
	}

	return $enc
}

# Return a xml string
proc ::feed::u::readXML { filename } {
	set enc [getEncoding $filename]
	set fd [open $filename]
	fconfigure $fd -encoding $enc
	set xml [read $fd]
	close $fd

	return $xml
}



namespace eval ::feed {
	# supported namespace prefixes
	# (FIXME: rely on namespaces not _prefixes_!
	# see http://groups.google.com/group/comp.text.xml/msg/5b9ca46272319bf5)
	variable sns [list itunes]
}

proc ::feed::init { Dom_doc filename } {
	upvar $Dom_doc doc

	set xml [::feed::u::readXML $filename]
	set doc [dom parse $xml]
	return [$doc documentElement]
}

# $node_obj must be a root element
proc ::feed::parse { node_obj Fdct } {
	upvar $Fdct f

	# If an attribute is namespaced (and matched by the pattern) or the
	# attribute is a namespace declaration, it is returned as a list
	# with three elements.
	# -- Rolf Ade, http://tech.groups.yahoo.com/group/tdom/message/316
	set ns ""
	set allns ""
	foreach i [$node_obj attributes] {
		if {[llength $i] != 3} continue
		lappend allns [lindex $i 0]
	}
	# extract only those prefixes that we can understand
	foreach i $allns {
		if {[lsearch -exact $::feed::sns $i] != -1} { lappend ns $i }
	}
	dict set f namespaces $ns

	set type [::feed::get_type $node_obj]
	switch -glob -- $type {
		atom-* {
			::feed::parse_atom $node_obj f
		}
		rss-* {
			::feed::parse_rss $node_obj f
		}
		default {
			error "unknown feed type \"$type\""
		}
	}

	::feed::verify $f
}

# $node_obj must be a root element
#
# return a feed type:
#    atom-1.0, rss-2.0, rss-2.1
#
# any other returned string is an error description
proc ::feed::get_type { node_obj } {

	set node_name [$node_obj nodeName]

	switch -- $node_name {
		feed {
			if {[$node_obj namespaceURI] ne "http://www.w3.org/2005/Atom"} {
				return "atom with unknown namespace"
			} else { return "atom-1.0" }
		}
		rss {
			set ver [$node_obj getAttribute version ""]
			if {[regexp -- {2\.0|2\.1} $ver]} {
				return rss-$ver
			}

			if {$ver == ""} {
				return "rss with missing version specification"
			} else {
				return "rss with unsupported version $ver"
			}
		}
		default {
			return "unknown root node \"$node_name\""
		}
	}
}

proc ::feed::touch_fdct_meta { Fdct } {
	upvar $Fdct f

	# NB, what is missing in our metadata, but exists in Atom spec:
	# icon, id, logo, rights, author->uri

	dict set f meta author name ""
	dict set f meta author email ""

	# a list from values of "term" attributes; "scheme" and "label"
	# attributes are ignored
	dict set f meta category ""; # a list

	dict set f meta contributor "";	# a list {{name email} {. .} ...}

	dict set f meta generator name ""
	dict set f meta generator version ""
	dict set f meta generator uri ""

	# only 1 and act as ref='alternate' -- this is a very simlified
	# solution and does not conform to Atom spec
	dict set f meta link ""

	dict set f meta subtitle ""; # analogue of rss <description>
	dict set f meta title ""
	dict set f meta updated "";	# analogue of rss <pubDate> and <lastBuildDate>
}

proc ::feed::touch_fdct_entry { Fdct n } {
	upvar $Fdct f

	# NB, what is missing in our entries, but exists in Atom spec:
	# rights, extensionElement, source, author->uri

	dict set f entries $n author name ""
	dict set f entries $n author email ""

	# a list from values of "term" attributes; "scheme" and "label"
	# attributes are ignored
	dict set f entries $n category ""; # a list

	dict set f entries $n content type "text"
	dict set f entries $n content text ""

	dict set f entries $n id ""
	dict set f entries $n contributor ""; # a list

	# only 1 and act as ref='alternate' -- this is a very simlified
	# solution and does not conform to Atom spec
	dict set f entries $n link ""

	# analogue to rss <enclosure> -- as <link rel="enclosure" .../> in Atom.
	# this is a list with nested list {{type} {length} {href}}
	dict set f entries $n enclosure ""

	dict set f entries $n published ""; # analogue of an rss item <pubDate>
	dict set f entries $n updated "";
	dict set f entries $n summary ""
	dict set f entries $n title ""
}

# Return:
#		big system-dependent integer value - OK
#		-1 -- error
proc ::feed::date_parse { date } {
	if {[regexp -- {^[A-Z][a-z]{2}, \d} $date]} {
		# suppose the date is in rfc822 format
		# (for example, Sun, 04 Nov 2007 18:08:00 +0200)
		set f "%a, %e %b %Y %T"
		if {[regexp -- {^[A-Z][a-z]{2}, \d+ \S+ \d+ \d\d:\d\d[^:]} $date]} {
			# some feed generators doesn't want to include seconds in the
			# date
			set f "%a, %e %b %Y %H:%M"
		}
		lappend f "$f %z"
	} else {
		# suppose the date is in ISO8601 format,
		# see http://en.wikipedia.org/wiki/ISO8601
		set f "%Y-%m-%dT%T"
		if {[regexp -- {^\d+-\d+-\d+T\d+:\d+:\d+(\.\d+)} $date match ms]} {
			# blogger.com form
			set f "%Y-%m-%dT%T${ms}"
		}
		lappend f "${f}%z"
	}

#	puts "****** HERE_IS_A_DATE_${date}_${f}"
	foreach i $f {
		if {![catch {clock scan $date -format $i -timezone :UTC} result]} {
			return $result
		}
	}

	# final try; "I gave up the attempt in despair"
	if {$date != "" && ![catch {clock scan $date -timezone :UTC} result]} {
		return $result
	}

	return -1
}

# make sure we are consuming rss safely by removing unsafe tags.
#
# $html -- an html string that we need to clean up
#
# return the cleaned-up html string
proc ::feed::html_remove_unsafe { html } {
	set unsafe_tags {
		script
		embed
		object
		frameset
		frame
		iframe
		meta
		link
		style
	}

	foreach tag $unsafe_tags {
		regsub -all "(<$tag\[^>\]*>(\[^<\]*</$tag>)?)+"	 $html {} html
	}

	return $html
}

#proc ::feed::get_node_text { node search {def ""}} {
#	if {[set n [$node selectNode "$search\[1]"]] != ""} {
#		return [string trim [$n text]]
#	} else { return $def }
#}

proc ::feed::get_node_text { node search {def ""}} {
	if {[set n [$node selectNode "*\[local-name()='$search']\[1]" ]] != ""} {
		return [atom_type_adaptation [$n getAttribute type ""] [$n text]]
	} else { return $def }
}

proc ::feed::get_sibling_nodes_text { node search {def ""}} {
#	puts "HERE $search"
	foreach i [$node selectNodes $search] {
		lappend r [string trim [$i text]]
	}
	if {[info exists r]} {
		return $r
	} else { return $def }
}

# atom style
proc ::feed::get_node_val { node search {def ""}} {
	foreach i [$node selectNodes $search ]  {
		# FIXME: replace asText with asHTML
		lappend r [atom_type_adaptation [$i getAttribute type ""] [$i asText]]
	}

	if {[info exists r]} { return $r }
	return $def
}

# atom style
proc ::feed::get_node_attr { node search attr {def ""}} {
	foreach i [$node selectNodes $search ]  {
		if {[set t [$i getAttribute $attr ""]] != ""} {
			lappend r [string trim $t]
		}
	}

	if {[info exists r]} { return $r }
	return $def
}

proc ::feed::get_one_of_flat_node_text { node search_list {def ""}} {
	foreach i $search_list {
		if {[llength [set r [::feed::get_sibling_nodes_text $node $i]]] != 0} {
			return $r
		}
	}
	return $def
}

# return a list {{name} {email}}
#
# $tags -- a list
proc ::feed::rss_get_author { node tags } {
	set r [list "" ""]
	# we extract only the first author
 	set author [lindex [::feed::get_one_of_flat_node_text $node $tags] 0]
 	if {$author != ""} {
 		# check if author is a email address
 		if {[regexp -- {(\S+@\S+\.\S+)(\s+\(.+\))?} $author match email name]} {
 			lset r 0 [string trim $name { \t)(}]
 			lset r 1 $email
 		} else {
 			lset r 0 $author
 		}
 	}

	return $r
}

proc ::feed::ns_meta_itunes { node Fdct } {
	upvar $Fdct fdct

	# 0. author
	if {[dict get $fdct meta author name] == ""} {
		set author [::feed::rss_get_author $node [list itunes:author]]
		dict set fdct meta author name [lindex $author 0]
		dict set fdct meta author email [lindex $author 1]
	}

	# 1. category
	set category [dict get $fdct meta category]
	foreach i [split [::feed::get_node_text $node itunes:keywords] ,] {
		if {![string is space $i]} { lappend category [string trim $i] }
	}
	dict set fdct meta category $category
}

proc ::feed::ns_entry_itunes { node Fdct n } {
	upvar $Fdct fdct

	# 0. author
	if {[dict get $fdct entries $n author name] == ""} {
		set author [::feed::rss_get_author $node [list itunes:author]]
		dict set fdct entries $n author name [lindex $author 0]
		dict set fdct entries $n author email [lindex $author 1]
	}

	# 2. content
	if {[dict get $fdct entries $n content text] == ""} {
		if {[set content [::feed::get_node_text $node \
									   itunes:summary] ] != ""} {
			dict set fdct entries $n content text \
				[::feed::html_remove_unsafe $content]
			dict set fdct entries $n content type html
		}
	}
}

# TODO: write something more intresting
proc ::feed::html_tide { t } {
	return [regsub -all -- {\s+} $t { }]
}

# Fill a dict $Fdct with 1 feed metadata and many feed entries.
proc ::feed::parse_rss { node_obj Fdct } {
	upvar $Fdct fdct

	set node [$node_obj firstChild]
	if {[$node nodeName] ne "channel"} {
		error "broken feed--no channel tag"
	}
	::feed::touch_fdct_meta fdct

	#
	# grab metadata
	#

	# 0. author
	set author [::feed::rss_get_author $node [list managingEditor]]
	dict set fdct meta author name [lindex $author 0]
	dict set fdct meta author email [lindex $author 1]

	# 1. category
	dict set fdct meta category [::feed::get_sibling_nodes_text $node category]

	# 2. contributor -- missing in rss

	# 3. generator
	dict set fdct meta generator name [::feed::get_node_text $node generator]

	# 4. link
	dict set fdct meta link [::feed::get_node_text $node link]

	# 5. subtitle
	dict set fdct meta subtitle \
		[::feed::html_tide [::feed::get_node_text $node description]]

	# 6. title
	dict set fdct meta title \
		[::feed::html_tide [::feed::get_node_text $node title]]

	# 7. updated
	if {[llength [set date [::feed::get_one_of_flat_node_text $node \
								[list lastBuildDate pubDate]] ] ] != 0} {
		dict set fdct meta updated [::feed::date_parse [lindex $date 0]]
	}

	# update all we did with additional parsing
	foreach i [dict get $fdct namespaces] {
		::feed::ns_meta_$i $node fdct
	}

	#
	# grab entries
	#

	set n 0
	foreach i [$node selectNode item] {
		::feed::touch_fdct_entry fdct $n

		# 0. author
		set author [::feed::rss_get_author $i [list author]]
		dict set fdct entries $n author name [lindex $author 0]
		dict set fdct entries $n author email [lindex $author 1]

		# 1. category
		dict set fdct entries $n category [::feed::get_sibling_nodes_text $i category]

		# 2. content
		if {[set content [::feed::get_node_text $i description] ] != ""} {
			dict set fdct entries $n content text \
				[::feed::html_tide [::feed::html_remove_unsafe $content]]
			# this is dumb I know, but don't blame me--blame rss 2.0 spec
			dict set fdct entries $n content type html
		}

		# 3. id
		dict set fdct entries $n id [::feed::get_node_text $i guid]

		# 4. contributor -- missing in rss

		# 5. link
		dict set fdct entries $n link [::feed::get_node_text $i link]

		# 6. enclosure
		set ee ""
		foreach e [$i selectNode enclosure] {
			lappend ee [list [$e getAttribute type ""] \
							[$e getAttribute length 0] \
							[$e getAttribute url ""]]
			# ouch
			for {set j 0} {$j < [llength [lindex $ee end]]} {incr j} {
				lset ee end $j [string trim [lindex [lindex $ee end] $j]]
			}
		}
		dict set fdct entries $n enclosure $ee

		# 7. published
		if {[set published [::feed::get_node_text $i pubDate]] != ""} {
			dict set fdct entries $n published [::feed::date_parse $published]
		}

		# 8. updated -- missing in rss
		# 9. summary -- missing in rss

		# 10. title
		dict set fdct entries $n title \
			[::feed::html_tide [::feed::get_node_text $i title]]

		# update all we did with additional parsing
		foreach j [dict get $fdct namespaces] {
			::feed::ns_entry_$j $i fdct $n
		}

		incr n
	}
}

# return empty string if a feed is OK,
# otherwise return some human description of the error.
proc ::feed::verify { fdct } {
	if {![dict exists $fdct entries]} {
		error "broken feed--cannot find any entries in it"
	}

	set ok 0
	set items [dict get $fdct entries]
	foreach i [dict keys $items] {
		foreach {name val} [dict get $items $i] {
			if {$name eq "enclosure" && [llength $val] != 0} {
				set ok 1
				break
			}
		}
	}

	if {!$ok} { error "cannot find enclosures, so it's not a podcast feed" }
	return ""
}

proc ::feed::atom_type_adaptation { type src } {
	switch -- $type {
		"" -
		"text" {
			return [string trim $src]
		}
		"html" -
		"xhtml" {
			# TODO: character entity references for ISO 8859-1 characters
			return [string trim [html_tide [html_remove_unsafe $src]]]
		}
		default {
			return $src
#			error "unknown value \"$type\" in Atom type attribute"
		}
	}
}

# return a list {{name email} {. .} ...}
proc ::feed::atom_person { node tag } {
	set r {}
	foreach i [$node selectNodes "*\[local-name()='$tag']" ] {
		set name ""
		set email ""

		foreach j [list name email] {
			if {[set n [$i selectNodes "*\[local-name()='$j']" ]] != ""} {
				set $j [string trim [$n text]]
			}
		}

		if {$name != "" || $email != ""} {
			lappend r [list $name $email]
		}
	}

	return $r
}

# Fill a dict $Fdct with 1 feed metadata and many feed entries.
proc ::feed::parse_atom { node_obj Fdct } {
	upvar $Fdct fdct

	::feed::touch_fdct_meta fdct

	#
	# grab metadata
	#

	# 0. author
	set author [lindex [atom_person $node_obj author] 0]
	dict set fdct meta author name [lindex $author 0]
	dict set fdct meta author email [lindex $author 1]

	# 1. category (unlike in RSS 2.0, useful data in Atom are
	#              being kept in attribute "term")
	dict set fdct meta category \
		[::feed::get_node_attr $node_obj {*[local-name()='category']} term]

	# 2. contributor
	dict set fdct meta contributor [atom_person $node_obj contributor]

	# 3. generator
	dict set fdct meta generator name \
		[lindex [get_node_val $node_obj {*[local-name()='generator']} ] 0]
	dict set fdct meta generator uri \
		[lindex [get_node_attr $node_obj {*[local-name()='generator']} uri] 0]
	dict set fdct meta generator version \
		[lindex [get_node_attr $node_obj {*[local-name()='generator']} version] 0]

	# 4. link (may be it's better to add condition @type='text/html'
	#    to XPath expression?)
	dict set fdct meta link \
		[lindex [::feed::get_node_attr $node_obj \
					 {*[local-name()='link' and @rel='alternate']} href] 0]

	# 5. subtitle
	dict set fdct meta subtitle \
		[lindex [get_node_val $node_obj {*[local-name()='subtitle']} ] 0]

	# 6.
	dict set fdct meta title \
		[lindex [get_node_val $node_obj {*[local-name()='title']} ] 0]

	# 7. updated
	if {[llength [set date [get_node_val $node_obj \
								{*[local-name()='updated']} ]]] != 0} {
		dict set fdct meta updated [::feed::date_parse [lindex $date 0]]
	}

	#
	# grab entries
	#


	set n 0
	foreach i [$node_obj selectNodes {*[local-name()='entry']} ] {
		::feed::touch_fdct_entry fdct $n

		# 0. author
		set author [lindex [atom_person $i author] 0]
		dict set fdct entries $n author name [lindex $author 0]
		dict set fdct entries $n author email [lindex $author 1]

		# 1. category
		dict set fdct entries $n category \
			[::feed::get_node_attr $i {*[local-name()='category']} term]

		# 2. content
		dict set fdct entries $n content text \
			[lindex [get_node_val $i {*[local-name()='content']} ] 0]
		dict set fdct entries $n content type \
			[lindex [get_node_attr $i {*[local-name()='content']} type] 0]

		# 3. id
		dict set fdct entries $n id \
			[lindex [get_node_val $i {*[local-name()='id']} ] 0]

		# 4. contributor
		dict set fdct entries $n contributor [atom_person $i contributor]

		# 5. link
		dict set fdct entries $n link \
			[lindex [get_node_attr $i \
						 {*[local-name()='link' and @rel='alternate']} href] 0]

		# 6. enclosure
		set ee ""
		foreach e [$i selectNodes {*[local-name()='link' and @rel='enclosure']} ] {
			lappend ee [list [$e getAttribute type ""] \
							[$e getAttribute length 0] \
							[$e getAttribute href ""]]
			# ouch
			for {set j 0} {$j < [llength [lindex $ee end]]} {incr j} {
				lset ee end $j [string trim [lindex [lindex $ee end] $j]]
			}
		}
		dict set fdct entries $n enclosure $ee

		# 7. published
		if {[llength [set date [get_node_val $i \
									{*[local-name()='published']} ]]] != 0} {
			dict set fdct entries $n published [::feed::date_parse [lindex $date 0]]
		}

		# 8. updated
		if {[llength [set date [get_node_val $i \
									{*[local-name()='updated']} ]]] != 0} {
			dict set fdct entries $n updated [::feed::date_parse [lindex $date 0]]
		}

		# 9. summary
		dict set fdct entries $n summary \
			[lindex [get_node_val $i {*[local-name()='summary']} ] 0]

		# 10. title
		dict set fdct entries $n title \
			[lindex [get_node_val $i {*[local-name()='title']} ] 0]

		incr n
	}
}

# $fdct -- a dict of a parsed feed
proc ::feed::dump { fdct } {
	puts "Feed metadata:"
	puts "Feed namespaces: [dict get $fdct namespaces]"
	foreach {i j} [dict get $fdct meta] {
		puts "* $i: $j"
	}

	puts "\nEntries of the feed:"
	set items [dict get $fdct entries]
	set total [dict size $items]
	foreach i [dict keys $items] {
		puts "\n$i/$total"

		foreach {name val} [dict get $items $i] {
			puts "* $name: $val"
		}
	}
}

# $order -- "increasing", "decreasing".
proc ::feed::sort { Fdct order } {
	upvar $Fdct fdct

	if {$order == ""} return

	# make a list {{entry_number date mark} {entry_number date mark} ...}
	set sort_entries ""

	set items [dict get $fdct entries]
	foreach i [dict keys $items] {
		set published [dict get $items $i published]
		if {$published == ""} { set published -1 }
		set updated [dict get $items $i updated]
		if {$updated == ""} { set updated -1 }

		set max $published
		if {$published < $updated} { set max $updated }
		lappend sort_entries [list $i $max 0]
	}

	veputs 4 "unsorted: $sort_entries\n"
	set sort_entries [lsort -integer -$order -index 1 $sort_entries]
	veputs 4 "  sorted: $sort_entries\n"

	# swap entries according to $sort_entries
	set n 0
	foreach i [dict keys $items] {
		set t_num [lindex [lindex $sort_entries $n] 0]
		if {$t_num == $i} {
			veputs 4 "entry #$i doesn't need to be sorted\n"
		} else {
			set t_val [dict get $fdct entries $i]

			if {[lindex [lindex $sort_entries $t_num] 2] == 0} {
				dict set fdct entries $i [dict get $fdct entries $t_num]
				dict set fdct entries $t_num $t_val

				# mark as swapped
				lset sort_entries $n 2 1
			} else {
				veputs 4 "entry #$i\($t_num) is already swapped\n"
			}
		}

		incr n
	}
	veputs 4 "  marked: $sort_entries\n"
}
