# Partial OPML parser for the opml-files from Google Reader.
#
# Example:
#
# set root [::opml::init dom_doc [open "my.xml"]]
#
# puts "opml version: [::opml::get_ver $root]"
#
# if {[catch {::opml::parse $root odct} cause]} {
# 	puts "$::errorInfo"; exit 1
# }
# ::opml::dump $odct
#
# $dom_doc delete; # free tDOM memory

package require tdom

namespace eval ::opml {
}

proc ::opml::init { Dom_doc fd } {
	upvar $Dom_doc doc

	# FIXME: this breaks on utf-8
	if {[catch {dom parse -channel $fd} doc]} {	error $doc }
	if {[::opml::get_ver [$doc documentElement]] == ""} {
		error "XML is not an OPML"
	}
	return [$doc documentElement]
}

# $node_obj must be a root element
#
# Return an OPML version (a string). An empty string is an error.
proc ::opml::get_ver { node_obj } {
	set node_name [$node_obj nodeName]

	if {$node_name eq "opml"} {
		set ver [$node_obj getAttribute version ""]
		if {[regexp -- {\s*\d+\.\d+\s*} $ver]} { return [string trim $ver] }
	}
	return ""
}

# $node_obj must be a root element.
#
# Fill Meta (a hash) with OPML metadata. Can raise an error.
proc ::opml::get_meta { node_obj Meta } {
	upvar $Meta meta

	set head_nodes [$node_obj selectNodes {/opml/*[local-name()='head']}]
	if {[llength $head_nodes] != 1} {
		error "OPML MUST have exact 1 head element"
	}

	array set meta {}
	foreach i [$head_nodes childNodes] {
		set meta([$i nodeName]) [string trim [$i text]]
	}

	if {![info exists meta(title)]} { error "No title in head" }
}

# A helper for [::opml::get_outlines]. Uses recursion.
proc ::opml::outlines { node } {
	set r []
	foreach i [$node childNodes] {
		if {[$i nodeName] ne "outline"} continue

		array set outline {}
		set outline(text) [string trim [$i getAttribute text ""]]
		set outline(xmlUrl) [string trim [$i getAttribute xmlUrl ""]]
		set outline(htmlUrl) [string trim [$i getAttribute htmlUrl ""]]

		# htmlUrl isn't essential for _our_ purpose
		if {$outline(text) != "" && $outline(xmlUrl) != ""} {
			lappend r [array get outline]
		} elseif {$outline(text) != ""} {
			# this is a header for a nested outlines

			# recursion!
			set sub [::opml::outlines $i]
			if {[llength $sub]} {
				lappend r [list "text" $outline(text) "nodes" $sub]
			}
		}
	}
	return $r
}

# $node_obj must be a root element.
#
# Return a list of outlines suitable for [array set ...]:
# { text "Name of the Podcast"
#   xmlUrl "http://example.org/feed.xml"
#   htmlUrl "http://example.org" }
#
# Outlines can be nested:
#
# { text "group 1"
#   "nodes" {
#             { text "Name of the Podcast #1"
#               xmlUrl "http://example.org/feed.xml"
#               htmlUrl "http://example.org" }
#             { text "Name of the Podcast #2"
#               xmlUrl "http://example.com/feed2.xml"
#               htmlUrl "http://example.com" }
#           }
# }
#
# Can raise an error.
proc ::opml::get_outlines { node_obj } {
	set bnodes [$node_obj selectNodes {/opml/*[local-name()='body']}]
	if {![llength $bnodes]} { error "OPML MUST have body element" }
	set body_node [lindex $bnodes 0]

	return [::opml::outlines $body_node]
}
