# Additional processing of enclosures after downloading them.

package require myutils
package require html_tricks

namespace eval ::treatment {
	variable treat [dict create]
}

proc ::treatment::is_empty {} {
	variable treat
	return [dict size $treat]
}

# return a file list
proc ::treatment::dir_snapshot { dir } {
	return [glob -nocomplain -directory $dir -types f -- *]
}

# return a difference between current directory state and the result
# from [dir_snapshot]
proc ::treatment::dir_diff { dir snapshot } {
	set r {}
	foreach i [glob -nocomplain -directory $dir -types f -- *] {
		if {[lsearch -exact $snapshot $i] == -1} {
			lappend r $i
		}
	}

	return $r
}

proc ::treatment::id3v1 { Feed_info fdct n } {
	upvar $Feed_info feed_info

	set r [dict create]
	dict set r _feed_name [file tail $feed_info(name)];		# comp.foobar_pro
	dict set r _fmt $feed_info(treatment);		# format string

	dict set r artist [dict get $fdct entries $n author name]
	dict set r album [dict get $fdct meta title]
	dict set r song [dict get $fdct entries $n title]

	if {[set date [dict get $fdct entries $n updated]] == ""} {
		set date [dict get $fdct entries $n published]
	}
	dict set r date $date

	dict set r year ""
	if {[dict get $r date] != ""} {
		dict set r year [clock format [dict get $r date] -format "%Y"]
	}

	if {[set comment [dict get $fdct entries $n summary]] == ""} {
		set comment [dict get $fdct entries $n content text]
	}
	dict set r comment $comment

	# remove html tags and parse character entity references for
	# ISO 8859-1 characters in comment
	foreach i [dict keys $r] {
		dict set r $i [regsub -all -- {[><|&]} \
						   [::html_tricks::mapEscapes \
								[regsub -all -- {</??.*?>} [dict get $r $i] ""]] .]
	}

	return $r
}

# $fnames -- the result from [dir_diff]
# $params -- the result from [id3v1]
proc ::treatment::add { fnames params } {
	variable treat

	# do nothing if user do not want to treat files from this feed
	if {[dict get $params _fmt] != "" && [llength $fnames] != 0} {
		dict set treat [dict get $params _feed_name] [lindex $fnames 0] $params
	}
}

proc ::treatment::reset {} {
	variable treat

	unset treat
	set treat [dict create]
}

# $fname -- a file name to treat
# $params -- a subset of dict from [id3v1]
#
# If verbose == 2, only prints possible executing command.
proc ::treatment::execute { fname params } {
	global conf env
	variable treat

	if {$conf(verbose)} {
		set redfi stdout;		# redirect file
	} else {
		if {[catch {open /dev/null WRONLY} redfi]} {
			errx 69 $redfi
		}
	}

	# treatment scripts can use this to construct a log file name
	set env(URANIACAST_LOGS) $conf(logs)
	set env(URANIACAST_FEED_CUR) [dict get $params _feed_name]

	array set arr [list \
				   "f" $fname \
				   "p" [dict get $params _feed_name] \
				   "a" [dict get $params artist] \
				   "A" [dict get $params album] \
				   "s" [dict get $params song] \
				   "y" [dict get $params year] \
				   "d" [dict get $params date] \
				   "c" [dict get $params comment] \
				  ]
	veputs 2 "  Command: [config_str_format [dict get $params _fmt] arr]\n"
	set r 1
	if {$conf(verbose) < 2} {
		if {[catch {
			exec -- {*}[config_str_format [dict get $params _fmt] arr] >&@ $redfi
		} cause]} {
			warnx "cannot treat [file tail $fname]: $cause"
			set r 0
		}
	}

	unset env(URANIACAST_LOGS)
	unset env(URANIACAST_FEED_CUR)
	return $r
}

# returns a list of statistics of treated files:
# { total treat failed }
proc ::treatment::do {} {
	variable treat
	global conf

	set stat [list 0 0 0];		# total, treat, failed
	veputs 1 "\nPreparing to treat "
	if {[::treatment::is_empty] == 0} {
		veputs 1 "0 feeds... Step ommited.\n"
		return $stat
	}
	veputs 1 "[::treatment::is_empty] feed(s)...\n"

	set podcast [dict get $treat]
	foreach i [dict keys $podcast] {
		veputs 1 "\nTreating $i file(s):\n"

		set n 1
		set items [dict get $treat $i]
		set total [dict size $items]
		foreach fname [dict keys $items] {
			if {$conf(signal)} break;

			# current/total
			veputs 1 "* File $n/$total \[[file tail $fname]]: "
			veputs 2 "\n  Artist:  [string range [dict get $items $fname artist] 0 64]\n"
			veputs 2 "  Album:   [string range [dict get $items $fname album] 0 64]\n"
			veputs 2 "  Song:    [string range [dict get $items $fname song] 0 64]\n"
			veputs 2 "  Year:    [dict get $items $fname year]\n"
			veputs 2 "  Date:    [dict get $items $fname date]\n"
			veputs 2 "  Comment: [string range [dict get $items $fname comment] 0 64]\n"

			if {[execute $fname [dict get $items $fname]]} {
				# number of successfully treated files
				lset stat 1 [expr [lindex $stat 1] + 1]
			} else {
				# number of failed treated files
				lset stat 2 [expr [lindex $stat 2] + 1]
			}

			incr n
			lset stat 0 [expr [lindex $stat 0] + 1]; # total
		}
	}

	return $stat
}
