#!/bin/sh

# This script was written with special guidelines for uraniacast because
# it is intended to run by uraniacast, although it may be used as a
# standalone program as well.
#
# See uraniacast_treatment(7).
#
# Requires: ffmpeg 0.10.x, exiftool
#
# Test:
#
# $ (export URANIACAST_LOGS=.; export URANIACAST_FEED_CUR=foo.bar; \
#   ./uraniacast_t_convert.sh -t mp3 01.ogg)
#
# Then check 'foo.bar/treatments/01.ogg.txt' file.

SCRIPT_PATH=$(dirname $(realpath $0))
. $SCRIPT_PATH/lib/bourne/helper.sh
. $SCRIPT_PATH/lib/bourne/progress_reporter.sh

conf_stream=a:0
conf_bitrate=56
conf_from=''
conf_to=''
conf_log=''
conf_delete=1
conf_duration=0

# $1 - file name
#
# Return a duration in seconds or nothing on error.
ffmpeg_getDuration()
{
	ffmpeg -i "$1" 2>&1 -vn | \
		awk '/Duration/ {
split(substr($2, 0, 2*3+2), time, ":");
printf "%d\n", time[1]*3600+time[2]*60+time[3]
}'

}

ffmpeg_convert()
{
	local line
	local ntimes=0

	ffmpeg -i "$1" -vn \
		-c:${conf_stream} libmp3lame \
		-b:${conf_stream} ${conf_bitrate}k \
		-ac:${conf_stream} 2 \
		-q:${conf_stream} 9 \
		-map_metadata 0:s:${conf_stream} \
		-y "$2" 2>&1 | tee $conf_log | \
		awk -v RS=
 '/time=/ {
raw=gensub(/.*time=([0-9:.]+).*/, "\\1", "");
split(substr(raw, 0, 2*3+2), time, ":");
printf "%d\n", time[1]*3600+time[2]*60+time[3];
fflush()
}' | {
		while read -r line
		do
			echo $line | progress_reporter_update $pr
			ntimes=$((ntimes+1))
		done

		[ $ntimes -eq 0 ] && return 1
		return 0
	}

	# the value of ntimes variable above was lost, because 'while read'
	# creates a subshell; so we group while loop with ntimes check
	return $?
}

usage()
{
	echo "Usage: ${0##*/} [-D] [-f format] -t format file" 1>&2
	exit 1
}

#
# Main
#

prerequisites ffmpeg exiftool || exit 1

while getopts t:f:Dd opt
do
	case $opt in
		t)
			conf_to=$OPTARG
			;;
		f)
			conf_from=$OPTARG
			;;
		D)
			conf_delete=0
			;;
		d)
			# a 'secret' CLO
			conf_duration=1
			;;
		*|h)
		usage
		;;
	esac
done
shift `expr $OPTIND - 1`

[ -z "$conf_to" ] && usage
[ -z "$1" ] && errx usage
conf_log=`treatment_log "$1"` || exit 1

[ -z "$conf_from" ] || \
	[ "`exiftool -T -mimetype $1 2>/dev/null`" == "$conf_from" ] || \
	errx "$1 is not an $conf_from file"

max=`ffmpeg_getDuration "$1"`
[ -z $max ] && errx "broken audio stream $conf_stream in $1"
[ $conf_duration -eq 1 ] && {
	echo $max
	exit 0
}

input_ext=${1##*.}
output=$2
[ -z "$output" ] && {
	output=${1%.*}.$conf_to
	[ "$1" = "$output" ] && output=${1%.*}.1.$conf_to
}

progress_reporter_begin "${input_ext}->$conf_to: "
pr=`progress_reporter_new 0 $max`

ffmpeg_convert "$1" "$output"
if [ $? -eq 0 ] ; then
	progress_reporter_end 'ok'
	echo ''
	[ $conf_delete -eq 1 ] && rm "$1"
	exit 0
else
	progress_reporter_end 'FAILED'
	warnx "see $conf_log for more info"
	exit 1
fi
